unit untTopicDownloadThread;

interface

uses
  Classes, SysUtils, Dialogs, MMSystem,
  IdBaseComponent,
  IdComponent, IdTCPConnection, IdTCPClient, IdHTTP,
  untHttpClient, untStreamTool, untTool,
  untBBSCore, untTopic, untGlobal, untOption,
  untHttp3;

type
  TTopicDownloadThread = class(TThread)
  private
    { Private 錾 }
    FParent         : TTopic;
    FNoDownload     : Boolean;
    FTopicStateType : TTopicStateType;
    FDatFile        : TextFile;
    FNewMessage     : TTopicMessage;
    FBuffer         : TMemoryStream;
    FBufferReader   : TStreamReader;
    FReadPosition   : int64;
    FMessageCount   : integer;
    FNoFirstLine    : Boolean;
    FDatValue       : string;
    procedure LoadLog;
    function ParseMsg(line: string; newmsg: Boolean): TTopicMessage;
    procedure StreamWrite(const Buff; Count : int64);
  public
    procedure   Execute; override;
    constructor Create(Parent: TTopic; NoDownload : Boolean);
    procedure   RaiseTopicStateChangeEvent;
    procedure   RaiseMessageReceivedEvent;
    procedure   FreeThread;
  end;

implementation

{ TTopicDownloadThread }

constructor TTopicDownloadThread.Create(Parent: TTopic; NoDownload : Boolean);
begin
  inherited Create(true);

  FParent := Parent;
  FParent.CountUp;  // ΉCountDown͎OĂ܂BӁI
  FNoDownload := NoDownload;
end;

procedure TTopicDownloadThread.Execute;
var
  logpath        : string;
  range          : integer;
  sr             : TStreamReader;
  line           : string;
  firstchar      : char;
  msgcount       : integer;
  parse          : TTopicMessage;
  contentsize    : integer;
  blnError       : boolean;
  I              : integer;
  blnBreak       : boolean;
  starttime      : integer;
  totalsecond    : string;
  Http           : TIdHttp;
  WriteStream    : TOnWriteStream;
  proxyitem      : TStringArray;
  proxyhost      : string;
  proxyport      : integer;
label
  ErrorFinish,
  Finish;
begin
  starttime := timeGetTime;

  inherited;

  FParent.IsDownloadingTopic := true;
  FParent.CheckWriteFolder;
  Http := nil;

  blnError := false;

  // Oǂݍ
  LoadLog();

  if (FParent.Datsize = 0) and (FParent.MessageList.Count > 0) then
  begin
    blnError := true;
    FParent.StatusText := '폜ځ[ŃOl܂悤ł';
    FParent.IsError    := true;
    FTopicStateType := stStatusText;
    Synchronize(RaiseTopicStateChangeEvent);
    goto ErrorFinish;
  end;

  if Terminated = true then goto ErrorFinish;

  if FNoDownload = false then
  begin

    FParent.StatusText :=  'ڑ';
    FTopicStateType := stStatusText;
    Synchronize(RaiseTopicStateChangeEvent);

    if Terminated = true then goto ErrorFinish;

    WriteStream := TOnWriteStream.Create();
    WriteStream.OnWrite := StreamWrite;

    FBuffer       := TMemoryStream.Create;
    FBufferReader := TStreamReader.Create(FBuffer);
    FReadPosition := 0;

    FMessageCount := FParent.GotMessageCount;
    FDatValue := '';

    // C^[lbgǂݍ
    Http := TIdHttp.Create(nil);
    Http.Request.UserAgent    := APP_USERAGENT;
    //Http.Request.CustomHeaders.Values['If-Modified-Since'] := FParent.LastModified;

    // proxy
    if gReadProxyUse = true then
      if gReadProxy <> '' then
      begin
        proxyitem := Split(gReadProxy, ':');
        if length(proxyitem) >= 2 then
          proxyport := StrToIntNeo(proxyitem[1])
        else
          proxyport := 8080;
        proxyhost := proxyitem[0];

        Http.ProxyParams.ProxyServer := proxyhost;
        Http.ProxyParams.ProxyPort   := proxyport;
      end;

    //if gReadProxyUse then Http.ProxyParams  HttpRequest.Proxy := gReadProxy
    if FParent.DatSize > 0 then
      range := FParent.DatSize - 1
    else
      range := 0;
    Http.Request.ContentRangeStart := range;

    FNoFirstLine := true;

    try
      Http.Get(FParent.DatUrl, WriteStream);
    except
    end;

    FBuffer.Free;
    WriteStream.Free;
    FBufferReader.Free;

    if ( 200 <= Http.ResponseCode) and (Http.ResponseCode < 300) then
    begin

      // DATTCYL^
      contentsize := Http.Response.ContentLength;
      if contentsize > 0 then
      if FParent.DatSize = 0 then
        FParent.DatSize := contentsize
      else
        FParent.DatSize := FParent.DatSize + contentsize - 1;

      // Last-ModifiedL^
      if blnBreak = false then
        FParent.LastModified := Http.Response.RawHeaders.Values['Last-Modified'];

      FParent.LastReadDate := DateTimeToStr(Now());
    end;

    FParent.NewMessageCount := FParent.MessageList.Count - FParent.GotMessageCount;
    FParent.GotMessageCount := FParent.MessageList.Count;

    if blnError = false then
    begin
      FParent.DownloadState := dsNone;
      FTopicStateType := stDownloadState;
      Synchronize(RaiseTopicStateChangeEvent);

      //totalsecond := IntToStr((timeGetTime() - starttime) div 1000) + 'b';

      if FParent.NewMessageCount = 0 then
        FParent.StatusText := 'VȂ'
      else
        FParent.StatusText := IntToStr(FParent.NewMessageCount) + '̃XM';
      FTopicStateType := stStatusText;
      Synchronize(RaiseTopicStateChangeEvent);
    end;

  end;

  // DATt@C̏o
  logpath := FParent.LocalDir + FParent.TopicId + '.dat';
  AssignFile(FDatFile, logpath);
  if FileExists(logpath) = false then ReWrite(FDatFile);
  Append(FDatFile);
  Write(FDatFile, FDatValue);
  CloseFile(FDatFile);

  goto Finish;

ErrorFinish:

  // VKMbZ[W폜
  {
  for I := 0 to NewMessageList.Count - 1 do
  begin
    FParent.MessageList.Remove(NewMessageList[I]);
    TTopicMessage(NewMessageList[I]).Free;
  end;
  }

Finish:

  FParent.SaveIdx();

  FParent.DownloadState := dsNone;
  FTopicStateType := stDownloadState;
  Synchronize(RaiseTopicStateChangeEvent);

  FParent.IsDownloadingTopic := false;

  Http.Free;

  //Synchronize(FreeThread);
  Terminate;
  Free;

end;

procedure TTopicDownloadThread.StreamWrite(const Buff; Count: int64);
var
  line    : string;
  parse   : TTopicMessage;
begin

  FBuffer.Seek(0, soFromEnd);
  FBuffer.Write(Buff, Count);

  FBuffer.Seek(FReadPosition, soFromBeginning);
  while FBufferReader.ReadLine(line) do
  begin
    if FNoFirstLine then      
    begin
      FNoFirstLine := false;
      if FParent.DatSize > 0 then
        if line = '' then
        begin

          FParent.DownloadState := dsProcessing;
          FTopicStateType := stDownloadState;
          Synchronize(RaiseTopicStateChangeEvent);

          FParent.StatusText := 'ڑ܂';
          FTopicStateType := stStatusText;
          Synchronize(RaiseTopicStateChangeEvent);

          continue;
        end else
        begin
          FParent.StatusText := '폜ځ[ŃOl܂悤ł';
          FParent.IsError    := true;
          FTopicStateType := stStatusText;
          Synchronize(RaiseTopicStateChangeEvent);
          exit;
        end;
    end;

    FDatValue := FDatValue + line + #10;
    FReadPosition := FBuffer.Position;
    Inc(FMessageCount);
    parse   := ParseMsg(line, true);
  end;

  Synchronize(RaiseMessageReceivedEvent);

  FBuffer.Seek(0, soFromEnd);

end;

{ --------------------------------------------------------
  ֐: LoadLog
  pr  : O̓ǂݍ
    : Ȃ
  ߂l: Ȃ
  l  : Ȃ
  ------------------------------------------------------ }

procedure TTopicDownloadThread.LoadLog;
var
  localloaded : boolean;
  I           : Integer;
  msg         : TTopicMessage;
  msgcount    : Integer;
  logpath     : string;
  logvalue    : string;
  lines       : TStringArray;
  parse       : TTopicMessage;
begin

  if FParent.NoNotice = true then
  begin

    // Cxgs
    for I := 0 to FParent.MessageList.Count - 1 do
    begin
      msg := TTopicMessage(FParent.MessageList[I]);
      FParent.NewReceivedMessage := msg;
      FNewMessage := msg;
      Synchronize(RaiseMessageReceivedEvent);
    end;
    FParent.NoNotice := false;
  end;
  msgcount := FParent.MessageList.Count;

  if FParent.IsLoadedLocalDat = false then
  begin

 	  // [Jǂݍ
    logpath  := FParent.LocalDir + FParent.TopicId + '.dat';
    logvalue := GetFile(logpath);
    lines    := Split(logvalue, #10);
    for I := 0 to Length(lines) - 1 do
    begin

      // p[X
      Inc(msgcount);
      parse := ParseMsg(lines[I], false);
      if parse = nil then
      begin
        FParent.IsError := true;
        FParent.StatusText := 'G[';
        FTopicStateType := stStatusText;
        Synchronize(RaiseTopicStateChangeEvent);
        exit;
      end;
      localloaded := true;
    end;
    FParent.IsLoadedLocalDat := true;
  end;
	FParent.Gotmessagecount := FParent.MessageList.Count;

  if localloaded = true then
  begin
    FParent.DownloadState := dsDatLoaded;
    FTopicStateType := stDownloadState;
    Synchronize(RaiseTopicStateChangeEvent)
  end;

  Synchronize(RaiseMessageReceivedEvent);

end;

{ --------------------------------------------------------
  ֐: ParseMsg
  pr  : bZ[Ws͂TTopicMessage𐶐
    : line
  @@@: msgnumber - bZ[WNo
  @@@: newmsg    - VbZ[W
  ߂l: TTopicMessage
  l  : Ȃ
  ------------------------------------------------------ }
function TTopicDownloadThread.ParseMsg(line: string; newmsg: Boolean): TTopicMessage;
var
  msg   : TTopicMessage;
  Items : TStringArray;
begin

  // TTopic̃bN͂̊֐̊Oōs

  // ACeɕ
  msg   := nil;
  Items := Split(line, ('<>'));
	if Length(Items) > 3 then
  begin

	  // bZ[WNX
    msg := TTopicMessage.Create();
    msg.PostName     := '<B>' + Items[0] +  '</B>';
    msg.PostEmail    := Items[1];
   	msg.RestStr      := Items[2];
    msg.Body         := Items[3] ;
		msg.IsNewMessage := newmsg;

		// bZ[Wzɒǉ
    FParent.AddNewMessage(msg);
    FParent.IsGotLog := true;
  end;

  if msg = nil then
    ShowMessage('ʹװ');

	result := msg;
end;

procedure TTopicDownloadThread.RaiseMessageReceivedEvent;
begin
  if Assigned(FParent.OnMessageReceived) then
    FParent.OnMessageReceived(FParent);
end;

procedure TTopicDownloadThread.RaiseTopicStateChangeEvent;
begin
  gBBSCore.RaiseTopicStateChangeEvent(FParent, FTopicStateType); 
end;

procedure TTopicDownloadThread.FreeThread;
begin
  FParent.CountDown;
end;

end.
